#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <ddraw.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include "../../compiler/compiler.h"
#include "../../trans/trans.h"
#include "../../clipping/clip2d.h"
#include "../../clipping/clip3d.h"
#include "engine.h"

#define TITLE "Dynamic 3D polygon engine"
#define NAME  "Dyn3DEng"

LPDIRECTDRAW            lpDD;            // DirectDraw object
LPDIRECTDRAWSURFACE     lpDDSPrimary;    // DirectDraw primary surface
HBITMAP                 OffScreen_hbm;   // Off-screen bitmap handle
HWND                    Screen_hwnd;     // Screen window handle
RECT                    Screen_rect;     // Screen RECT
HLOCAL                  HVirtScr;        // Handle to Off-screen buffer

static void DestroyObjects(void)
{
       if (lpDD!=NULL)
       {
           if (lpDDSPrimary!=NULL)
           {
              lpDDSPrimary->Release();
              lpDDSPrimary =NULL;
           }
           lpDD->Release();
           lpDD =NULL;
       }
}// DestroyObjects

BOOL DirectDrawFail(HWND hwnd)
{
    DestroyObjects();
    MessageBox(hwnd,"DirectDraw Inititialization fail",TITLE,MB_OK);
    DestroyWindow(hwnd);
    return(FALSE);
} // DirectDrawFail

long FAR PASCAL WindowProc( HWND   hwnd  , UINT msg     ,
                            WPARAM wParam, LPARAM lParam  )
{
    switch (msg)
    {
//           case WM_CREATE:
//           break;
           case WM_SETCURSOR:
                SetCursor(NULL);
                return(TRUE);
           case WM_KEYDOWN:
                switch (wParam)
                {
                      case VK_ESCAPE:
                      case VK_F12:
                           PostMessage(hwnd,WM_CLOSE,0,0);
                      break;
//                      default:
                }
           break;
           case WM_DESTROY:
                PostQuitMessage(0);
           break;
    }
    return(DefWindowProc(hwnd,msg,wParam,lParam));
} // WindowProc

static BOOL InitializeApp(HINSTANCE hInstance,int nCmdShow)
{
    HWND                hwnd;
    WNDCLASS            wc;
    HRESULT             ddrval;
    DDSURFACEDESC       ddsd;
    HDC                 hdc;
    // set up and register window class
    wc.style          = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc    = WindowProc;
    wc.cbClsExtra     = 0;
    wc.cbWndExtra     = 0;
    wc.hInstance      = hInstance;
    wc.hIcon          = LoadIcon(hInstance,IDI_APPLICATION);
    wc.hCursor        = LoadCursor(NULL,IDC_ARROW);
    wc.hbrBackground  = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName   = NAME;
    wc.lpszClassName  = NAME;
    RegisterClass(&wc);
    // create a window
    hwnd=CreateWindowEx(
                        0,
                        NAME,
                        TITLE,
                        WS_POPUP,
                        0,
                        0,
                        GetSystemMetrics(SM_CXSCREEN),
                        GetSystemMetrics(SM_CYSCREEN),
                        NULL,
                        NULL,
                        hInstance,
                        NULL
                        );
    if (!hwnd)return(FALSE);
    Screen_hwnd=hwnd;
    ShowWindow(hwnd,nCmdShow);
    UpdateWindow(hwnd);

    // create direct draw objects
    ddrval=DirectDrawCreate(NULL,&lpDD,NULL);
    if (ddrval!=DD_OK)return(DirectDrawFail(hwnd));
    ddrval=lpDD->SetCooperativeLevel(hwnd,DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);
    if (ddrval!=DD_OK)return(DirectDrawFail(hwnd));
    ddrval=lpDD->SetDisplayMode(320,200,16);
    if (ddrval!=DD_OK)return(DirectDrawFail(hwnd));
    ddsd.dwSize        =sizeof(DDSURFACEDESC);
    ddsd.dwFlags       =DDSD_CAPS;
    ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE;
    ddrval=lpDD->CreateSurface(&ddsd,&lpDDSPrimary,NULL);
    if (ddrval!=DD_OK)return(DirectDrawFail(hwnd));
    if (GetWindowRect(Screen_hwnd,&Screen_rect)==0)return(DirectDrawFail(hwnd));
    // create off-screen bitmap and off-screen virtual screen
    hdc=GetDC(Screen_hwnd);
    OffScreen_hbm=CreateCompatibleBitmap(hdc,Screen_rect.right,Screen_rect.bottom);
    ReleaseDC(Screen_hwnd,hdc);
    HVirtScr=LocalAlloc(LMEM_MOVEABLE|LMEM_ZEROINIT,VirtualScrSize);
    VirtScr=(ColorVal *)LocalLock(HVirtScr);
    if (VirtScr==NULL)return(DirectDrawFail(Screen_hwnd));
    return(TRUE);
} // InitializeApp

void OS_BlitToScreen(void)
{
    HDC                 hdc_Source;
    HDC                 hdc_Dest;
//    DWORD               CurScan;

    lpDDSPrimary->Restore();
    hdc_Source=CreateCompatibleDC(NULL);
    SelectObject(hdc_Source,OffScreen_hbm);
    SetBitmapBits(OffScreen_hbm,VirtualScrSize,VirtScr);
    lpDDSPrimary->GetDC(&hdc_Dest);
/*    lpDD->GetScanLine(&CurScan);
    while (CurScan)
          lpDD->GetScanLine(&CurScan);*/
    BitBlt(hdc_Dest,0,0,320,200,hdc_Source,0,0,SRCCOPY);
    lpDDSPrimary->ReleaseDC(hdc_Dest);
    DeleteDC(hdc_Source);
}

void DestroyApp(void)
{
    LocalFree(HVirtScr);
    DeleteObject(OffScreen_hbm);
}

int PASCAL WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR     lpCmdLine, int       nCmdShow       )
{
    MSG                 msg;
    BOOL                ExitMessage=FALSE;

    lpCmdLine     = lpCmdLine;
    hPrevInstance = hPrevInstance;

    if (!InitializeApp(hInstance,nCmdShow))return(FALSE);
    if (!Engine_Init())return(FALSE);

//    short ssw;         // turn on infinity
//    _asm fstcw [ssw]
//    ssw&=(WORD)65535-(WORD)4096;
//    ssw|=(WORD)4096;
//    _asm fldcw [ssw]

    while (!ExitMessage)
    {
          if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
          {
             if (msg.message==WM_QUIT)ExitMessage=TRUE;
             TranslateMessage(&msg);
             DispatchMessage(&msg);
          }
          else
          {
             Engine_UpdateDisplay();
             OS_BlitToScreen();
          }
    }
    Engine_Destroy();
    DestroyApp();
    return(msg.wParam);
} // WinMain
